import edu.ufl.digitalworlds.j4k.J4KSDK;

import javax.swing.*;
import java.awt.*;
import java.util.HashMap;
import java.util.LinkedList;

public class Camera implements Runnable {

    PosCont curpos = null;

    JFrame wcframe;
    JFrame wcframe2;

    Kinect kinect;
    VideoPanel main_panel;
    VideoPanel debug_panel;
    byte[] image, obrazok, obrazok2;
    float[] labImage;
    int width = 640;
    int height = 480;

    CameraData cData;

    public Camera(CameraData cDataOut) {
        cData = cDataOut;
    }

    public void Init() {
        kinect = new Kinect();
        if (!kinect.start(J4KSDK.COLOR)) {
            System.err.println("Neda sa spustit kinect ...");
            System.exit(1);
        }
        main_panel = new VideoPanel();
        kinect.setCamera(this);

        wcframe = new JFrame("Kinect");
        wcframe.add(main_panel);
        wcframe.setDefaultCloseOperation(JFrame.DO_NOTHING_ON_CLOSE);
        wcframe.pack();
        wcframe.setSize(640, 480);
        wcframe.setVisible(true);

        debug_panel = new VideoPanel();

        wcframe2 = new JFrame("Debug");
        wcframe2.add(debug_panel);
        wcframe2.setDefaultCloseOperation(JFrame.DO_NOTHING_ON_CLOSE);
        wcframe2.pack();
        wcframe2.setSize(640, 480);
        wcframe2.setVisible(true);
    }

    synchronized public void setImageFromKinect(byte[] img, int w, int h) {
        image = img;
        width = w;
        height = h;
    }

    //ziskaj najaktualnejsiu poziciu
    public synchronized PosCont getPositions() {
        return new PosCont(curpos);
    }

    //set pos
    private synchronized void setPositions(PosCont pos) {
        curpos = pos;
    }

    private synchronized void copyImage() {
        obrazok = image.clone();
    }

    public synchronized void copyImage2() {
        obrazok2 = image.clone();
    }


    private void drawImage(byte[] img) {
        if (main_panel.videoTexture != null) {
            main_panel.videoTexture.update(width, height, img);
        }

    }

    private void setPixel(byte[] img, int x, int y, byte r, byte g, byte b) {
        if (x < 0 || y < 0 || y >= height || x >= width) return;
        int pos = (x + y * width) * 4;
        img[pos] = b;
        img[pos + 1] = g;
        img[pos + 2] = r;
    }

    // RGB z BGR
    private int[] getPixel(byte[] img, int x, int y) {
        if (x < 0 || y < 0 || y >= height || x >= width) return null;
        int[] o = new int[3];

        int pos = (x + y * width) * 4;
        o[0] = img[pos + 2] & 0xFF;
        o[1] = img[pos + 1] & 0xFF;
        o[2] = img[pos] & 0xFF;

        return o;
    }

    private void getPixel(float[] img, int x, int y, float[] out) {
        if (x < 0 || y < 0 || y >= height || x >= width) {
            out[0] = -1;
            out[1] = -200;
            out[2] = -200;
            return;
        }

        if (out.length < 3)
            out = new float[3];
        int pos = (x + y * width) * 3;
        out[0] = img[pos]; // tu to bolo naopak ...
        out[1] = img[pos + 1];
        out[2] = img[pos + 2];
    }

    private void convertToLAB() {
        labImage = new float[3 * width * height];
        for (int x = 0; x < width; x++) {
            for (int y = 0; y < height; y++) {
                int[] RGB = getPixel(obrazok, x, y);
                float[] rgbf = new float[3];
                if (RGB != null) {
                    rgbf[0] = (float) RGB[0] / 255f;
                    rgbf[1] = (float) RGB[1] / 255f;
                    rgbf[2] = (float) RGB[2] / 255f;
                }
                int pos = (x + y * width) * 3;
                float[] lab = CIELab.fromRGB(rgbf);
                labImage[pos] = lab[0];
                labImage[pos + 1] = lab[1];
                labImage[pos + 2] = lab[2];
            }
        }
    }

    private void getPixelFromCenter(float[] out) {
        float[] sum = new float[3];
        float[] temp = new float[3];
        int cnt = 0;
        for (int x = width / 2 - 3; x <= width / 2 + 3; x++) {
            for (int y = height / 2 - 3; y <= height / 2 + 3; y++) {
                getPixel(labImage, x, y, temp);
                sum[0] += temp[0];
                sum[1] += temp[1];
                sum[2] += temp[2];
                cnt++;
            }
        }
        sum[0] = sum[0] / (float) cnt;
        sum[1] = sum[1] / (float) cnt;
        sum[2] = sum[2] / (float) cnt;
        out[0] = sum[0];
        out[1] = sum[1];
        out[2] = sum[2];
    }

    private int[] prahuj(float[] from, float[] to) {
        int[] o = new int[width * height];
        float[] px = new float[3];
        float minL = Math.min(from[0], to[0]);
        float maxL = Math.max(from[0], to[0]);
        float minA = Math.min(from[1], to[1]);
        float maxA = Math.max(from[1], to[1]);
        float minB = Math.min(from[2], to[2]);
        float maxB = Math.max(from[2], to[2]);
        for (int x = 0; x < width; x++)
            for (int y = 0; y < height; y++) {
                getPixel(labImage, x, y, px);
                if (px[0] <= maxL && px[0] >= minL && px[1] <= maxA && px[1] >= minA && px[2] <= maxB && px[2] >= minB) {
                    o[y * width + x] = 1;
                } else {
                    o[y * width + x] = 0;
                }

            }
        return o;
    }

    void mask(int[] mask) {
        for (int x = 0; x < width; x++)
            for (int y = 0; y < height; y++) {
                if (mask[x + y * width] == 0) {
                    if (!(x >= width / 2 - 3 && x <= width / 2 + 3 && y >= height / 2 - 3 && y <= height / 2 + 3)) {
                        int[] pixel = getPixel(obrazok, x, y);
                        if (pixel != null) {
                            byte Y = (byte) (((pixel[0] * 0.2126f) + (pixel[1] * 0.7152f) + (pixel[2] * 0.0722f)) / 4);
                            setPixel(obrazok, x, y, Y, Y, Y);
                        }
                    }
                }
            }
        for (int x = width / 2 - 4; x <= width / 2 + 4; x++) {
            setPixel(obrazok, x, height / 2 - 4, (byte) 255, (byte) 255, (byte) 255);
            setPixel(obrazok, x, height / 2 + 4, (byte) 255, (byte) 255, (byte) 255);
        }
        for (int y = height / 2 - 4; y <= height / 2 + 4; y++) {
            setPixel(obrazok, width / 2 - 4, y, (byte) 255, (byte) 255, (byte) 255);
            setPixel(obrazok, width / 2 + 4, y, (byte) 255, (byte) 255, (byte) 255);
        }
    }

    private int[] drawX(int[] in) {
        int[] stred = new int[6];

        long sx = 0, sy = 0;
        int maxX = 0, maxY = 0, minX = width, minY = height;
        int pocet = 0;
        for (int x = 0; x < width; x++)
            for (int y = 0; y < height; y++) {
                if (in[y * width + x] > 0) {
                    sx += x;
                    sy += y;
                    pocet++;
                    if (x > maxX) maxX = x;
                    if (y > maxY) maxY = y;
                    if (x < minX) minX = x;
                    if (y < minY) minY = y;
                }
            }
        if (pocet == 0) return stred;

        stred[0] = Math.round(sx / pocet);
        stred[1] = Math.round(sy / pocet);
        stred[2] = minX;
        stred[3] = minY;
        stred[4] = maxX;
        stred[5] = maxY;

        //draw X
        drawSquare(stred);
        return stred;
    }

    public void drawSquare(int[] stred, byte[] color) {
        for (int x = -5; x < 5; x++)
            for (int y = -5; y < 5; y++) {
                setPixel(obrazok, stred[0] + x, stred[1] + y, color[0], color[1], color[2]);
                setPixel(obrazok, stred[0] - x, stred[1] + y, color[0], color[1], color[2]);
            }
    }

    public void drawSquare2(int[] stred, byte[] color) {
        for (int x = -5; x < 5; x++)
            for (int y = -5; y < 5; y++) {
                setPixel(obrazok2, stred[0] + x, stred[1] + y, color[0], color[1], color[2]);
                setPixel(obrazok2, stred[0] - x, stred[1] + y, color[0], color[1], color[2]);
            }
    }

    public void drawSquare(int[] stred) {
        drawSquare(stred, new byte[]{(byte) 255, (byte) 255, (byte) 255});
    }

    private void najSpoj(int[] in) {
        int N = 2;
        HashMap<Integer, Integer> score = new HashMap<>();

        int k;
        for (int x = 0; x < width; x++)
            for (int y = 0; y < height; y++) {
                if (in[x + y * width] == 1) {
                    k = flood(x, y, in, N);
                    score.put(N, k);
                    N++;
                }
            }
        int best = 2, bestVal = 0;
        int pom;
        //zistii ktora je naj
        for (Integer q : score.keySet()) {
            pom = score.get(q);
            if (bestVal < pom) {
                bestVal = pom;
                best = q;
            }
        }
        //sprav genocidu
        for (int x = 0; x < width; x++)
            for (int y = 0; y < height; y++) {
                if (in[x + y * width] != best) {
                    in[x + y * width] = 0;
                } else {
                    in[x + y * width] = 1;
                }
            }
    }

    private void gaussBlur(int r, byte[] img) {
        byte[] red = separateColor(2, img);
        byte[] green = separateColor(1, img);
        byte[] blue = separateColor(0, img);

        byte[] red2 = new byte[red.length];
        byte[] green2 = new byte[green.length];
        byte[] blue2 = new byte[blue.length];

        gaussBlurAlg(red, red2, width, height, r);
        gaussBlurAlg(green, green2, width, height, r);
        gaussBlurAlg(blue, blue2, width, height, r);

        putColor(2, red2, img);
        putColor(1, green2, img);
        putColor(0, blue2, img);
    }


    private void gaussBlurAlg(byte[] scl, byte[] tcl, int w, int h, int r) {
        int[] bxs = boxesForGauss(r, 3);
        boxBlur(scl, tcl, w, h, (bxs[0] - 1) / 2);
        boxBlur(tcl, scl, w, h, (bxs[1] - 1) / 2);
        boxBlur(scl, tcl, w, h, (bxs[2] - 1) / 2);
    }

    private void boxBlur(byte[] scl, byte[] tcl, int w, int h, int r) {
        System.arraycopy(scl, 0, tcl, 0, scl.length);
        boxBlurH(tcl, scl, w, h, r);
        boxBlurT(scl, tcl, w, h, r);
    }

    private void boxBlurH(byte[] scl, byte[] tcl, int w, int h, int r) {
        double iarr = 1 / (double) (r + r + 1);
        for (int i = 0; i < h; i++) {
            int ti = i * w, li = ti, ri = ti + r;
            double fv = scl[ti] & 255, lv = scl[ti + w - 1] & 255, val = (r + 1) * fv;
            for (int j = 0; j < r; j++) val += scl[ti + j] & 255;
            for (int j = 0; j <= r; j++) {
                val += (scl[ri++] & 255) - fv;
                tcl[ti++] = (byte) (Math.round(val * iarr));
            }
            for (int j = r + 1; j < w - r; j++) {
                val += (scl[ri++] & 255) - (scl[li++] & 255);
                tcl[ti++] = (byte) (Math.round(val * iarr));
            }
            for (int j = w - r; j < w; j++) {
                val += lv - (scl[li++] & 255);
                tcl[ti++] = (byte) (Math.round(val * iarr));
            }
        }
    }

    private void boxBlurT(byte[] scl, byte[] tcl, int w, int h, int r) {
        double iarr = 1 / (double) (r + r + 1);
        for (int i = 0; i < w; i++) {
            int ti = i, li = ti, ri = ti + r * w;
            double fv = scl[ti] & 255, lv = scl[ti + w * (h - 1)] & 255, val = (r + 1) * fv;
            for (int j = 0; j < r; j++) val += scl[ti + j * w] & 255;
            for (int j = 0; j <= r; j++) {
                val += (scl[ri] & 255) - fv;
                tcl[ti] = (byte) (Math.round(val * iarr));
                ri += w;
                ti += w;
            }
            for (int j = r + 1; j < h - r; j++) {
                val += (scl[ri] & 255) - (scl[li] & 255);
                tcl[ti] = (byte) (Math.round(val * iarr));
                li += w;
                ri += w;
                ti += w;
            }
            for (int j = h - r; j < h; j++) {
                val += lv - (scl[li] & 255);
                tcl[ti] = (byte) (Math.round(val * iarr));
                li += w;
                ti += w;
            }
        }
    }

    private int[] boxesForGauss(double sigma, int n)  // standard deviation, number of boxes
    {
        double wIdeal = Math.sqrt((12 * sigma * sigma / n) + 1);  // Ideal averaging filter width
        int wl = (int) Math.floor(wIdeal);
        if (wl % 2 == 0) wl--;
        int wu = wl + 2;

        double mIdeal = (12 * sigma * sigma - n * wl * wl - 4 * n * wl - 3 * n) / (-4 * wl - 4);
        int m = (int) Math.round(mIdeal);

        int[] sizes = new int[n];
        for (int i = 0; i < n; i++)
            sizes[i] = (i < m ? wl : wu);
        return sizes;
    }


    private byte[] separateColor(int offset, byte[] img) {
        byte[] output = new byte[width * height];

        if (img.length < width * height * 4) return output;

        for (int x = 0; x < width; x++)
            for (int y = 0; y < height; y++) {
                output[x + y * width] = img[(y * width + x) * 4 + offset];
            }

        return output;
    }

    private void putColor(int color, byte[] colors, byte[] trgt) {
        if (trgt.length < width * height * 4) return;
        for (int x = 0; x < width; x++)
            for (int y = 0; y < height; y++) {
                trgt[(y * width + x) * 4 + color] = colors[x + y * width];
            }
    }

    private int flood(int px, int py, int[] in, int n) {
        LinkedList<Point> q = new LinkedList<>();

        q.push(new Point(px, py));
        int x, y;
        Point p;
        int res = 0;
        while (!q.isEmpty()) {
            p = q.poll();
            x = p.x;
            y = p.y;
            if (in[x + y * width] != 1) {
                continue;
            }
            in[x + y * width] = n;
            res++;
            if (x + (y - 1) * width >= 0 && in[x + (y - 1) * width] == 1) {//up
                q.push(new Point(x, y - 1));
            }
            if (x + (y + 1) * width < in.length && in[x + (y + 1) * width] == 1) {//dwn
                q.push(new Point(x, y + 1));
            }
            if (x - 1 + (y) * width >= 0 && in[x - 1 + y * width] == 1) {//lft
                q.push(new Point(x - 1, y));
            }
            if (x + 1 + (y) * width < in.length && in[x + 1 + y * width] == 1) {//rght
                q.push(new Point(x + 1, y));
            }
            if (x - 1 + (y - 1) * width >= 0 && x - 1 + (y - 1) * width < in.length && in[x - 1 + (y - 1) * width] == 1) { //uplft
                q.push(new Point(x - 1, y - 1));
            }
            if (x + 1 + (y - 1) * width >= 0 && x + 1 + (y - 1) * width < in.length && in[x + 1 + (y - 1) * width] == 1) { //uprght
                q.push(new Point(x + 1, y - 1));
            }
            if (x - 1 + (y + 1) * width >= 0 && x - 1 + (y + 1) * width < in.length && in[x - 1 + (y + 1) * width] == 1) { //dwnlft
                q.push(new Point(x - 1, y + 1));
            }
            if (x + 1 + (y + 1) * width >= 0 && x + 1 + (y + 1) * width < in.length && in[x + 1 + (y + 1) * width] == 1) { //dwnrght
                q.push(new Point(x + 1, y + 1));
            }
        }
        return res;
    }

    @Override
    public void run() {
        while (true) {
            //nacitaj z kamerky obrazok
            float[] stred = new float[3];
            if (image != null && image.length > 0) {
                copyImage();
                gaussBlur(3, obrazok);
                convertToLAB();
                getPixelFromCenter(stred);
            } else {
                continue;
            }
            int[] mask = new int[width * height];

            int[] processed_orange = prahuj(cData.getOrange_from(), cData.getOrange_to());
            najSpoj(processed_orange);
            mergeMask(mask, processed_orange);

            int[] process_green = prahuj(cData.getGreen_from(), cData.getGreen_to());
            najSpoj(process_green);
            mergeMask(mask, process_green);

            int[] process_blue = prahuj(cData.getBlue_from(), cData.getBlue_to());
            najSpoj(process_blue);
            mergeMask(mask, process_blue);

            mask(mask);
            int[] ball = drawX(processed_orange);
            int[] robfront = drawX(process_green);
            int[] robback = drawX(process_blue);


            //vypocitaj pozicie a narvi ich do curpos
            PosCont cmpPos = new PosCont();
            cmpPos.valid = 1;
            cmpPos.Lopta[0] = ball[0];
            cmpPos.Lopta[1] = ball[1];
            cmpPos.Ciel[0] = width / 2;
            cmpPos.Ciel[1] = height / 2;
            cmpPos.polomerLopty = ((ball[3] - ball[1]) / 2);
            int c2 = ((ball[2] - ball[0]) / 2);
            if (c2 > cmpPos.polomerLopty)
                cmpPos.polomerLopty = c2;
            if (1 > cmpPos.polomerLopty)
                cmpPos.polomerLopty = 1;
            cmpPos.polomerLopty = 9;
            cmpPos.Robot[0] = (robfront[0] + robback[0]) / 2;
            cmpPos.Robot[1] = (robfront[1] + robback[1]) / 2;

            cmpPos.RobAngle = PosCont.calcHeading(robback, robfront);
            setPositions(cmpPos);
            drawSquare(cmpPos.Ciel, new byte[]{(byte) 255, (byte) 255, (byte) 64});
            if (Thread.currentThread().isInterrupted()) {
                break;
            }

            wcframe.setTitle("Kinect: {" + stred[0] + ", " + stred[1] + ", " + stred[2] + "}");
            drawImage(obrazok);

            Main.notifyMe();
        }
    }

    private void mergeMask(int[] mask, int[] processed) {
        for (int x = 0; x < width; x++)
            for (int y = 0; y < height; y++) {
                int pos = x + y * width;
                mask[pos] = mask[pos] | processed[pos];
            }
    }

    public synchronized void drawDataImg() {
        if (obrazok == null) return;
        if (obrazok2 == null)
            obrazok2 = new byte[obrazok.length];
        debug_panel.videoTexture.update(width, height, obrazok2);
    }
}
