import org.apache.commons.math3.fitting.PolynomialCurveFitter;
import org.apache.commons.math3.fitting.WeightedObservedPoints;

import java.awt.*;
import java.io.*;
import java.util.HashSet;
import java.util.Set;


public class Regres {

    Set<Point> regStrela;
    Set<Point> regFwd;
    Set<Point> regRot;

    public Regres() {
        regStrela = new HashSet<>();
        regFwd = new HashSet<>();
        regRot = new HashSet<>();
    }

    int getStrel(int distance) {
        //vrati silu pre strelu ktora ma ist distance
        return doRegres(regStrela, distance, 2);
    }

    int getRotation(int angle) {
        //vrati kolko sa ma robot tocit aby sa otocil o angle
        return doRegres(regRot, angle, 2);
    }

    int getFwd(int distance) {
        //vrati kolko ma robot ist aby presiel distance
        return doRegres(regFwd, distance, 1);
    }

    public void putStrel(int pwr, PosCont before, PosCont after) {
        regStrela.add(new Point(pwr, (int) Math.round(PosCont.calcDist(before.Lopta, after.Lopta))));
    }

    public void putFwd(int t, PosCont before, PosCont after) {
        regFwd.add(new Point(t, (int) Math.round(PosCont.calcDist(before.Robot, after.Robot))));
    }

    public void putRotation(int t, PosCont before, PosCont after) {

        int d = Math.abs(before.RobAngle - after.RobAngle) % 360;
        int r = d > 180 ? 360 - d : d;
        regRot.add(new Point(t, r));
    }


    private int doRegres(Set<Point> a, int query, int polyDeg) {//spravi regresiu dat a vrati hodnotu fcie pre query

        if (a.size() < 1)
            return query;

        WeightedObservedPoints obs = new WeightedObservedPoints();
        for (Point p : a) {
            obs.add(p.x, p.y);
        }

        PolynomialCurveFitter fitter = PolynomialCurveFitter.create(polyDeg);

        // Retrieve fitted parameters (coefficients of the polynomial function).
        double[] coeff = fitter.fit(obs.toList());
        double res = 0;

        for (int i = 0; i < coeff.length; i++) {
            //System.out.println(coeff[i]+" x^"+i);
            res += Math.pow(query, i) * coeff[i];
        }


        return (int) Math.round(res);
    }

    //napis do suboru
    public void save(String meno) {
        System.out.println("save invoked at save invocation point");
        try {
            FileOutputStream fos = new FileOutputStream(new File(meno));
            ObjectOutputStream oos;
            oos = new ObjectOutputStream(fos);
            oos.writeObject(regStrela);
            oos.writeObject(regFwd);
            oos.writeObject(regRot);
            oos.close();
        } catch (Exception e) {
            System.out.println("Save sa pokazil oprav alebo skusaj donekonecna");
        }

    }

    //nacitaj zo suboru
    @SuppressWarnings("unchecked")
    public void load(String meno) {
        System.out.println("loaaaaaadingering.....");
        try {
            FileInputStream fis = new FileInputStream(new File(meno));
            ObjectInputStream ois = new ObjectInputStream(fis);
            regStrela = (Set<Point>) ois.readObject();
            regFwd = (Set<Point>) ois.readObject();
            regRot = (Set<Point>) ois.readObject();
            ois.close();
        } catch (Exception e) {
            System.out.println("Load sa pokazil je zly format mozno");
            e.printStackTrace();
        }
        System.out.println("------NAUCENIE-------");
        System.out.println("strel" + regStrela.size());
        printCoefs(regStrela);
        System.out.println("fwd" + regFwd.size());
        printCoefs(regFwd);
        System.out.println("rot" + regRot.size());
        printCoefs(regRot);
        System.out.println("------end-------");
        for (Point p : regRot) {
            System.out.println(p.toString());
        }
        System.out.println("------end3-------");
    }

    private void printCoefs(Set<Point> a) {

        if (a.size() < 1)
            return;
        WeightedObservedPoints obs = new WeightedObservedPoints();
        for (Point p : a) {
            obs.add(p.x, p.y);
        }

        //instancujme polynom nteho stupna
        int n = 3;
        PolynomialCurveFitter fitter = PolynomialCurveFitter.create(n);

        // Retrieve fitted parameters (coefficients of the polynomial function).
        double[] coeff = fitter.fit(obs.toList());
        double res = 0;

        for (int i = 0; i < coeff.length; i++) {
            System.out.println(coeff[i] + " x^" + i);
        }


    }

    public boolean test() {
        //3x^3 + 2x^2 + 11x +4
        Set<Point> mmm = new HashSet<Point>();
        mmm.add(new Point(7, 1208));
        mmm.add(new Point(5, 484));
        mmm.add(new Point(12, 5608));
        mmm.add(new Point(32, 100708));
        mmm.add(new Point(-3, -92));
        mmm.add(new Point(-7, -1004));
        mmm.add(new Point(-15, -9836));
        mmm.add(new Point(0, 4));
        int res = doRegres(mmm, 10, 3);//malo by sa rovnat3314
        System.out.println("odpoved " + res + " a malo by byt 3314");
        return true;
    }
}
